/***************************************************************************
  par UTPECA d'après Rui Santos
  Complete project details at
  https://RandomNerdTutorials.com/esp32-async-web-server-espasyncwebserver-library/
  https://RandomNerdTutorials.com/display-images-esp32-esp8266-web-server/
  https://randomnerdtutorials.com/esp-now-esp32-arduino-ide/
  https://randomnerdtutorials.com/esp32-esp-now-wi-fi-web-server/ 
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

   Croquis émetteur pour le Calcul d'itinéraire :
   fonctionne avec les croquis récepteurs
     - TCObyUTPECA4b1 pour la commande des appareils
     - TCObyUTPECA4c1 pour l'alimentation des sections
****************************************************************************/

/*--------------------------------------------------------------------------
   Import required libraries
  --------------------------------------------------------------------------*/
#ifdef ESP32
#include <esp_now.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <SPIFFS.h>
#else
#include <Arduino.h>
#include <espnow.h>
#include <ESP8266WiFi.h>
#include <Hash.h>
#include <ESPAsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "FS.h"
#endif

/*--------------------------------------------------------------------------
   DONNEES DE LA TACHE TCO_WEB
  --------------------------------------------------------------------------*/
// Connection au routeur wifi (la box)
#define DEBUG
const char* ssid     = "... le nom de votre routeur ...";
const char* password = "... votre mot de passe ...";
      bool  b_Appr   = false;           // pour modifier appareil voie
      bool  b_Alim   = false;           // pour alimenter itinéraire

// communication entre le serveur et la page HTML (browser)
const char*  PARAM_INPUT_1 = "appareil"; // Nom de l'appareil
const char*  PARAM_INPUT_2 = "state";    // Etat de l'appareil
const char*  PARAM_INPUT_3 = "section";  // Nom de la section
      String inputMessage1;
      String inputMessage2;
      String inputMessage3;

/*--------------------------------------------------------------------------
   DEFINITION DE LA TABLE DES RECEPTEURS
  --------------------------------------------------------------------------*/
typedef struct
{
  bool    actif;
  char    tyRcp;                         // A = appareil, K = section
  uint8_t bcAdr[6];
} recepteur;

// Table des RECEIVER'S MAC ADDRESS
#define NBMAC 4
recepteur t_Mac[NBMAC] =
{
  // mettre ici les adresses MAC de vos microprocesseurs
/*{ false, 'A', {0x0C, 0xB8, 0x15, 0xF7, 0x9E, 0xE4}},      // appareils
  { false, 'A', {0x10, 0x97, 0xBD, 0xE4, 0xBF, 0x64}},      
  { false, 'S', {0xAC, 0x0B, 0xFB, 0xD9, 0x8A, 0xA0}},      // sections
  { false, 'S', {0x8A, 0xA0, 0xAC, 0x0B, 0xFB, 0xD9}} */      
  { false, 'A', {0xAC, 0x0B, 0xFB, 0xD9, 0x8A, 0xA0}},      // appareils
  { false, 'A', {0x10, 0x97, 0xBD, 0xE4, 0xBF, 0x64}},      
  { false, 'K', {0x0C, 0xB8, 0x15, 0xF7, 0x9E, 0xE4}},      // cantons
  { false, 'K', {0x8A, 0xA0, 0xAC, 0x0B, 0xFB, 0xD9}}       
};

/*--------------------------------------------------------------------------
   DONNEES PROTOCOLE ESP-NOW
  --------------------------------------------------------------------------*/
struct comm
{
  char   typRcp;       // type de récepteur (A = appareil, K = section)
  String nomKod;       // nom de l'appareil ou de la section

  union
  {
    struct             // Appareil de voie
    {
      byte    numApp;  // numéro appareil (de 1 à 8)
      char    typApp;  // type de l'appareil ('A', 'X', 'T', 'Y')
      char    etaApp;  // état de l'appareil ( de '1' à '4')
      byte    noMot1;  // No du 1er moteur (de 1 à 8)
      byte    noMot2;  // No du 2me moteur (de 1 à 8)
    };
    struct             // Section de voie
    {
      char    etaSek;  // état : 0->Off, 1->alimenté
      byte    no_Rel;  // No du relais (de 1 à 8)
      byte    id_Rcp;  // No récepteur (rang+1 des macAdr)
      byte    dummy1;
      byte    dummy2;
    };
  };
};

//Create a struct_message called myComm
struct comm myComm;

// gestion des appariements
#ifdef ESP32
esp_now_peer_info_t peerInfo;
#endif

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);


// Prise en compte du fichier source du TCO
/*********************************************************************/
#define TCO1
#ifdef TCO1
#include "Combrailles.h"
#endif
#ifdef TCO2
#include "Saint-Sernin.h"
#endif
#ifdef TCO4
#include "Mon_TCO_No04.h"
#endif

/*********************************************************************/
void CommandeAppareil(const String& nomTyp, const String& etaApp)
/*********************************************************************/
{
  int    colPos = nomTyp.indexOf(':');
  String nomApp = nomTyp.substring(0, colPos);   // nom de l'appareil
  String typApp = nomTyp.substring(colPos + 1);  // type de l'appareil (Aig, Tjd, ...
  String numApp = nomTyp.substring(3, colPos);   // numéro de l'appareil
  int    no_App = numApp.toInt() - 1;            // ...en indice de table
  t_App[no_App].etaApp = etaApp[0];              // maj état de l'appareil

  myComm.typRcp = 'A';                           // A = appareil de voie
  myComm.nomKod = nomApp;
  myComm.numApp = no_App;
  myComm.typApp = typApp[0];
  myComm.etaApp = etaApp[0];
  myComm.noMot1 = t_App[no_App].noMot1;
  myComm.noMot2 = t_App[no_App].noMot2;
  int    noMac  = int(t_App[no_App].noRcpt) - 1;

  // Debug : affichage des appareils et de leur état
  //===================================================================
  char macStr[18];
  Serial.print   ( nomApp );
  Serial.print   ( ", Etat" );
  Serial.print   ( etaApp );
  Serial.print   ( " ==> Mac[");
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           t_Mac[noMac].bcAdr[0], t_Mac[noMac].bcAdr[1], t_Mac[noMac].bcAdr[2],
           t_Mac[noMac].bcAdr[3], t_Mac[noMac].bcAdr[4], t_Mac[noMac].bcAdr[5]);
  Serial.print(macStr);
  Serial.print   ( "] : ");

  if ((noMac >= 0) && (t_Mac[noMac].actif))
  {
#ifdef ESP32
    esp_err_t result1 = esp_now_send(t_Mac[noMac].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
    (result1 == ESP_OK) ? Serial.print("Envoi OK " ) : Serial.print ("Envoi KO ");
#else
    esp_now_send(t_Mac[noMac].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
    Serial.print("Envoi OK " );
#endif
    delay(500);
  }
  else
  {
    Serial.println("Pas de récepteur");
  }
}

/*********************************************************************/
int ChercheSection( String& nom)
/*********************************************************************/
{
  int j = -1;
  for (int i=0; i<NBSEK; i++)
  {
    if (t_Sek[i].nomSek == nom)
    {
      j = i;
      break;
    }
  }
  return(j);
}

/*********************************************************************/
void AlimenteSection( String& nom, const String& eta)
/*********************************************************************/
{
  int    colPos = 0;
  int    b_Fin  = 1;
  int    num    = -1;
  int    noMac  = 0;
  String strSek = nom;
  String nomSek = "";
  
  while (b_Fin > 0)
  {
    colPos        = strSek.indexOf(':');
    if (colPos < 0)
    {
      b_Fin  = 0;
      nomSek = strSek;                              // nom de la section
    }
    else
    {
      nomSek = strSek.substring(0, colPos);         // nom de la section
    }
    num           = ChercheSection(nomSek);         // rang de la section
    if (num < 0) { continue; }
    noMac         = int(t_Sek[num].noRcpt) - 1;
    t_Sek[num].etaSek = eta[0];                     // maj état de la section

    myComm.typRcp = 'K';                            // K = section de voie
    myComm.nomKod = nomSek;
    myComm.etaSek = eta[0];
    myComm.no_Rel = t_Sek[num].no_Rel;

    // Debug : affichage des sections et de leur état
    //===================================================================
    char macStr[18];
    Serial.print   ( t_Sek[num].nomSek );
    Serial.print   ( ", Etat" );
    Serial.print   ( t_Sek[num].etaSek );
    Serial.print   ( " ==> Mac[");
    snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
             t_Mac[noMac].bcAdr[0], t_Mac[noMac].bcAdr[1], t_Mac[noMac].bcAdr[2],
             t_Mac[noMac].bcAdr[3], t_Mac[noMac].bcAdr[4], t_Mac[noMac].bcAdr[5]);
    Serial.print(macStr);
    Serial.print   ( "] : ");

    if ((noMac >= 0) && (t_Mac[noMac].actif))
    {
#ifdef ESP32
      esp_err_t result1 = esp_now_send(t_Mac[noMac].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
      (result1 == ESP_OK) ? Serial.print("Envoi OK " ) : Serial.print ("Envoi KO ");
#else
      esp_now_send(t_Mac[noMac].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
      Serial.print("Envoi OK " );
#endif
      delay(500);
    }
    else
    {
      Serial.println("Pas de récepteur");
    }
    if (colPos >= 0)
    { strSek = strSek.substring(colPos + 1); }  // le reste de la liste des cantons
  }
}

/*********************************************************************/
void RAZSection( String& nom, const String& eta)
/*********************************************************************/
{
  for (int i = 0; i < NBMAC; i++)
  {
    if (t_Mac[i].tyRcp == 'K') 
    {
      myComm.typRcp = 'K';                       // K = section de voie
      myComm.nomKod = nom;
      myComm.etaSek = eta[0];
      myComm.no_Rel = 0;
      t_Sek[i].etaSek = eta[0];                  // maj état de la section

      // Debug : affichage des sections et de leur état
      //===================================================================
      char macStr[18];
      Serial.print   ( nom );
      Serial.print   ( ", Etat" );
      Serial.print   ( eta );
      Serial.print   ( " ==> Mac[");
      snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
             t_Mac[i].bcAdr[0], t_Mac[i].bcAdr[1], t_Mac[i].bcAdr[2],
             t_Mac[i].bcAdr[3], t_Mac[i].bcAdr[4], t_Mac[i].bcAdr[5]);
      Serial.print(macStr);
      Serial.print   ( "] : ");

      if (t_Mac[i].actif)
      {
#ifdef ESP32
        esp_err_t result1 = esp_now_send(t_Mac[i].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
        (result1 == ESP_OK) ? Serial.print("Envoi OK " ) : Serial.print ("Envoi KO ");
#else
        esp_now_send(t_Mac[i].bcAdr,  (uint8_t *) &myComm, sizeof(comm));
        Serial.print("Envoi OK " );
#endif
        delay(500);
      }
      else
      {
        Serial.println("Pas de récepteur");
      }
    }
  }
}

#ifdef ESP32
/*********************************************************************/
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status)
/*********************************************************************/
// callback when data is sent
{
  Serial.println(status == ESP_NOW_SEND_SUCCESS ? "==> Livraison acceptée" : "==> Livraison refusée");
}
#else
/*********************************************************************/
void OnDataSent(uint8_t *mac_addr, uint8_t status)
/*********************************************************************/
// callback when data is sent
{
  Serial.println(status == 0 ? "==> Livraison acceptée" : "==> Livraison refusée");
}
#endif

/*********************************************************************/
void InitialiseAppareils()
/*********************************************************************/
{
  //  String nomTyp = "";

  for (int i = 0; i < NBMOT; i++)
  {
    if (t_App[i].typApp != ' ')
    {
      String etaApp = String(t_App[i].etaApp);
      String nomTyp = "App" + String(i + 1) + ":" + String(t_App[i].typApp);
      if (t_App[i].noRcpt > 0)
      {
        CommandeAppareil(nomTyp, etaApp);
      }
    }
  }
}

/*********************************************************************/
void InitialiseSections()
/*********************************************************************/
{
  String allRaz = "RAZ_";
  RAZSection(allRaz, String('0'));
}

/*********************************************************************/
void setup()
/*********************************************************************/
{
  // Serial port for debugging purposes
  //------------------------------------------------------------------
  Serial.begin(115200);
  Serial.print("### TCO-Web/Emetteur (v4.00) sur ");
#ifdef ESP32
  Serial.println("ESP32 ###");
#else
  Serial.println("ESP8266 ###");
#endif
  Serial.print("### TCO installé : ");
  Serial.print(nomTCO);
  Serial.println(" ###");

  // Connection à ESP-NOW
  //------------------------------------------------------------------
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  Serial.print("Initialisation ESP-NOW : ");
#ifdef ESP32
  if (esp_now_init() != ESP_OK) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_register_send_cb(OnDataSent);
#else
  if (esp_now_init() != 0) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_set_self_role(ESP_NOW_ROLE_CONTROLLER);
  esp_now_register_send_cb(OnDataSent);
#endif

  // Appairage des récepteurs
  //------------------------------------------------------------------
  char macStr[18];
  for (int noMac = 0; noMac < NBMAC; noMac++)
  {
    snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
             t_Mac[noMac].bcAdr[0], t_Mac[noMac].bcAdr[1], t_Mac[noMac].bcAdr[2],
             t_Mac[noMac].bcAdr[3], t_Mac[noMac].bcAdr[4], t_Mac[noMac].bcAdr[5]);
    Serial.print("Appairage avec Mac[");
    Serial.print(macStr);
    Serial.print("] : ");

#ifdef ESP32
    peerInfo.channel = 0;
    peerInfo.encrypt = false;
    // register peer
    memcpy(peerInfo.peer_addr, t_Mac[noMac].bcAdr, 6);
    if (esp_now_add_peer(&peerInfo) != ESP_OK)
    {
      Serial.println("Echec");
      t_Mac[noMac].actif = false;
    }
    else
    {
      Serial.println("Ok");
      t_Mac[noMac].actif = true;
    }
#else
    if (esp_now_add_peer(t_Mac[noMac].bcAdr,  ESP_NOW_ROLE_SLAVE, 1, NULL, 0) != 0)
    {
      Serial.println("Echec");
      t_Mac[noMac].actif = false;
    }
    else
    {
      Serial.println("Ok");
      t_Mac[noMac].actif = true;
    }
#endif
  }

  // Initialisation SPIFFS
  //------------------------------------------------------------------
  if (!SPIFFS.begin())
  {
    Serial.println("Initialisation SPIFFS incorrecte");
    return;
  }
  else
  {
    Serial.println("Initialisation SPIFFS : OK");
  }

  // Connection à Wi-Fi
  //------------------------------------------------------------------
  Serial.print   ( "Initialisation Wifi " );
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  // Print ESP Local Data
  Serial.println ( "" );
  Serial.print   ( "Maintenant connecté à " );
  Serial.println ( ssid );
  Serial.print   ( "Adresse IP : " );
  Serial.println ( WiFi.localIP() );
  Serial.print   ( "mac Adress : " );
  Serial.println ( WiFi.macAddress() );
  Serial.print   ( "WiFi channel : " );
  Serial.println ( WiFi.channel() );

  // Démarrage serveur web
  //------------------------------------------------------------------
  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request)
    { request->send(SPIFFS, "/index.html", String(), false, processor); });

  ChargerVignettes();

  // Send a GET request to <ESP_IP>/update?appareil=<inputMessage1>&state=<inputMessage2>
  server.on("/update", HTTP_GET, [] (AsyncWebServerRequest * request) 
  {
    // GET input1 value on <ESP_IP>/update?appareil=<inputMessage1>&state=<inputMessage2>
    if (request->hasParam(PARAM_INPUT_1) && request->hasParam(PARAM_INPUT_2))
    {
      inputMessage1 = request->getParam(PARAM_INPUT_1)->value();
      inputMessage2 = request->getParam(PARAM_INPUT_2)->value();
      // pour contrer le watchdog de Async_tcp :
      // on déporte le travail à faire ds fonction loop
      b_Appr        = true;    
    }
    else if (request->hasParam(PARAM_INPUT_3) && request->hasParam(PARAM_INPUT_2))
    {
      inputMessage3 = request->getParam(PARAM_INPUT_3)->value();
      inputMessage2 = request->getParam(PARAM_INPUT_2)->value();
      // pour contrer le watchdog de Async_tcp
      // on déporte le travail à faire ds fonction loop
      b_Alim        = true;    
    }
    else
    {
      inputMessage1 = "Message ";
      inputMessage2 = "KO";
      Serial.print(inputMessage1);
      Serial.println(inputMessage2);
    }

    request->send(200, "text/plain", "OK");
  });

  // Start server
  server.begin();

  // Initialisation Appareils
  //------------------------------------------------------------------
  Serial.println ( "Début initialisation appareils " );
  InitialiseAppareils();
  Serial.println ( "Initialisation appareils terminée " );

  // Initialisation Sections de voie
  //------------------------------------------------------------------
  Serial.println ( "Début initialisation sections de voie " );
  InitialiseSections();
  Serial.println ( "Initialisation sections de voie terminée " );
}

/*********************************************************************/
void loop()
/*********************************************************************/
{
  // pour contrer le watchdog de Async_tcp
  // le travail à faire est déporté ici
  if (b_Appr)
  {
    CommandeAppareil(inputMessage1, inputMessage2);
    b_Appr = false;
  }
  
  if (b_Alim)
  {
    if (inputMessage3.substring(0, 3) == "RAZ")
    { RAZSection(inputMessage3, inputMessage2); }
    else
    { AlimenteSection(inputMessage3, inputMessage2); }
    b_Alim = false;
  }
}
